@@ -2,18 +2,17 @@
 console application environment
 ===============================
 
-the :class:`ConsoleApp` allows your application the easy declaration of command line arguments and options.
+an instance of the :class:`ConsoleApp` class is representing a python application with dynamically configurable logging,
+debugging features (inherited from :class:`~ae.core.AppBase`), command line arguments and config files and options.
 
-:class:`ConsoleApp` inherits from the :class:`~ae.core.AppBase` application base class, providing dynamically
-configurable logging and debugging features (see also the :mod:`docstrings of the core module <ae.core>`).
+the helper function :func:`sh_exec` provided by this portion simplifies the execution of shell/console commands.
 
 
 define command line arguments and options
 -----------------------------------------
 
-after creating an instance of the class :class:`ConsoleApp`, use the methods :meth:`~ConsoleApp.add_argument` and
-:meth:`~ConsoleApp.add_option` to define the command line arguments and the :ref:`config options <config-options>`.
-finally call the :meth:`~ConsoleApp.run_app` method to parse the command line arguments::
+the methods :meth:`~ConsoleApp.add_argument` and :meth:`~ConsoleApp.add_option` are defining command line arguments and
+:ref:`config options <config-options>`, finally parsed/loaded by calling :meth:`~ConsoleApp.run_app`::
 
     ca = ConsoleApp(app_title="command line arguments demo", app_version="3.6.9")
     ca.add_argument('argument_name_or_id', help="Help text for this command line argument")
@@ -21,18 +20,20 @@ finally call the :meth:`~ConsoleApp.run_app` method to parse the command line ar
     ...
     ca.run_app()
 
-the values of the commend line arguments can be determined by calling the methods :meth:`~ConsoleApp.get_argument` and
-:meth:`~ConsoleApp.get_option` of the :class:`ConsoleApp` app instance. additional configuration values, persistently
-stored in :ref:`INI/CFG files <config-files>`, are provided by the :meth:`~ConsoleApp.get_variable` method.
+the values of the commend line arguments and options are determined via the methods :meth:`~ConsoleApp.get_argument` and
+:meth:`~ConsoleApp.get_option`. additional configuration values, stored in :ref:`INI/CFG files <config-files>`, are
+accessible via the :meth:`~ConsoleApp.get_variable` method.
 
 
-auto-collecting features
-------------------------
+auto-collected app name, title and version
+------------------------------------------
 
 .. _app-title:
 .. _app-version:
 
-this code example, a skeleton of an app module, would run just fine - without raising any `AssertionError`::
+if one of the kwargs :paramref:`~ConsoleApp.app_title` or :paramref:`~ConsoleApp.app_version` is not specified in the
+init call of the :class:`ConsoleApp` class instance, then they will automatically get determined from your app main
+module: the app title from the docstring title, and the application version string from the `__version__` variable::
 
     \"\"\" module docstring title \"\"\"
     from ae.console import ConsoleApp
@@ -44,15 +45,11 @@ this code example, a skeleton of an app module, would run just fine - without ra
     assert ca.app_title == "module docstring title"
     assert ca.app_version == '1.2.3'
 
-if one of the kwargs :paramref:`~ConsoleApp.app_title` or :paramref:`~ConsoleApp.app_version` is not specified in the
-init call of the instance `ca`, then it automatically collects the app title from the docstring title of the module, and
-the application version string from the module variable `__version__`.
-
 .. _app-name:
 
 :class:`ConsoleApp` also determines on instantiation the name/id of your application, if not explicitly specified in
 :paramref:`~ConsoleApp.app_name`. other application environment vars/options (like e.g. the application startup folder
-path and the current working directory path) will be automatically initialized and provided via the `ca` instance.
+path and the current working directory path) will be automatically initialized and provided via the app instance.
 
 
 configuration files, sections, variables and options
@@ -108,7 +105,7 @@ config file with two config sections containing one config option (named `log_fi
 
     [aeOptions]
     log_file = './logs/your_log_file.log'
-    configVar1 = ['list-element1', ('list-element2-1', 'list-element2-2', ), dict()]
+    configVar1 = ['list-element1', ('list-element2-1', 'list-element2-2', ), {}]
 
     [YourSectionName]
     configVar2 = {'key1': 'value 1', 'key2': 2222, 'key3': datetime.datetime.now()}
@@ -135,20 +132,17 @@ see :attr:`special encapsulated strings <ae.literal.Literal.value>`, respectivel
 the following config variables are pre-defined in the :ref:`main config section <config-main-section>` and recognized by
 :mod:`this module <.console>`, some of them also by the module/portion :mod:`ae.core`:
 
-* `debug_level` : debug logging verbosity level (this is also a :ref:`config option <config-options>` - set-able as
-  command line arg).
-* `logging_params` : general logging configuration parameters (py and ae logging)
-  - :meth:`documented here <.core.AppBase.init_logging>`.
-* `py_logging_params` : configuration parameters to activate python logging -
-  `documented in the Python docs <https://docs.python.org/3.6/library/logging.config.html#logging.config.dictConfig>`_.
-* `log_file` : log file name for ae logging (this is also a :ref:`config option <config-options>` - set-able as command
-  line arg).
-* `onboarding_tour_started` : count the onboarding tour starts since the installation of the app. will be reset to zero
-  after a user registration (by calling :meth:`~ConsoleApp.register_user`).
-* `registered_users` : users registered with their OS user name as user id (see :meth:`~ConsoleApp.register_user`).
-* `user_id` : id of the app user (default is determined from the `system user name <ae.base.os_user_name>`)
-* `user_specific_cfg_vars` : list of config variables storing an individual value for each registered user (see
-  section :ref:`user-specific-config-variables`).
+* `debug_level`: debug logging verbosity level :ref:`config option <config-options>`
+* `log_file`: ae logging file name (this is also a :ref:`config option <config-options>` - set-able as command line arg)
+* `logging_params`: :meth:`general ae logging configuration parameters (py and ae logging) <.core.AppBase.init_logging>`
+* `py_logging_params`: `python logging configuration
+  <https://docs.python.org/3.6/library/logging.config.html#logging.config.dictConfig>`_
+* `onboarding_tour_started`: count the onboarding tour starts since the installation of the app. will be reset to zero
+  after a user registration (by calling :meth:`~ConsoleApp.register_user`)
+* `registered_users`: users registered with their OS user name as user id (see :meth:`~ConsoleApp.register_user`)
+* `user_id`: id of the app user (default is determined from the `system user name <ae.base.os_user_name>`)
+* `user_specific_cfg_vars`: list of config variables storing an individual value for each registered user (see
+  section :ref:`user-specific-config-variables`)
 
 .. note::
   the value of a config variable can be overwritten by defining an OS environment variable with a name that is equal to
@@ -196,7 +190,6 @@ be abbreviated on the command line with the short `-L` option id.
     after an explicit definition of the optional config option `user_id` via :meth:`~ConsoleApp.add_option` it will be
     automatically used to initialize the :attr:`~ConsoleApp.user_id` attribute.
 
-
 .. _user-specific-config-variables:
 
 user specific config variables
@@ -213,23 +206,25 @@ revoke which config variables the app is storing individually for each user.
 """
 import os
 import datetime
+import shlex
+import subprocess
 import threading
 
-from typing import Any, Callable, Dict, Iterable, Optional, Set, Type, Tuple
+from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Set, Type, Tuple, Union
 from configparser import ConfigParser, NoSectionError
 from argparse import ArgumentParser, ArgumentError, HelpFormatter, Namespace
 
-from ae.base import (  # type: ignore
-    CFG_EXT, DATE_TIME_ISO, DATE_ISO, INI_EXT, UNSET,
-    env_str, instantiate_config_parser, norm_name, os_user_name, sys_env_dict, sys_env_text)
-from ae.paths import norm_path, Collector, PATH_PLACEHOLDERS            # type: ignore
+from ae.base import (                                                                       # type: ignore
+    CFG_EXT, DATE_TIME_ISO, DATE_ISO, INI_EXT, UnsetType, UNSET,
+    dummy_function, env_str, instantiate_config_parser, norm_name, os_user_name, sys_env_dict, sys_env_text)
+from ae.paths import normalize, Collector, PATH_PLACEHOLDERS                                # type: ignore
 # noinspection PyProtectedMember
-from ae.core import (                                                   # type: ignore  # for mypy
+from ae.core import (                                                                       # type: ignore  # for mypy
     DEBUG_LEVEL_DISABLED, DEBUG_LEVELS, main_app_instance, ori_std_out, _LOGGER, AppBase)
-from ae.literal import Literal                                          # type: ignore
+from ae.literal import Literal                                                              # type: ignore
 
 
-__version__ = '0.2.54'
+__version__ = '0.3.66'
 
 
 MAIN_SECTION_NAME: str = 'aeOptions'            #: default name of main config section
@@ -256,6 +251,46 @@ def config_value_string(value: Any) -> str:
     return str_val.replace('%', '%%')
 
 
+def sh_exec(command_line: str, extra_args: Sequence = (), console_input: str = "",
+            lines_output: Optional[List[str]] = None, cae: Optional[Any] = None, shell: bool = False) -> int:
+    """ execute command in the current working directory of the OS console/shell.
+
+    :param command_line:        command line string to execute on the console/shell. could contain command line args
+                                separated by whitespace characters (alternatively use :paramref:`~sh_exec.extra_args`).
+    :param extra_args:          optional sequence of extra command line arguments.
+    :param console_input:       optional string to be sent to the stdin stream of the console/shell.
+    :param lines_output:        optional list to return the lines printed to stdout/stderr on execution.
+    :param cae:                 optional :class:`~ae.console.ConsoleApp` instance, only used for logging. to suppress
+                                any logging output pass :data:`~ae.base.UNSET`.
+    :param shell:               pass True to execute command in the default OS shell (see :meth:`subprocess.run`).
+    :return:                    return code of the executed command or 126 if execution raised any other exception.
+    """
+    # args = command_line + " " + " ".join(extra_args) if shell else command_line.split() + list(extra_args)
+    args = command_line + " " + " ".join(extra_args) if shell else shlex.split(command_line) + list(extra_args)
+    print_out = cae.po if cae else print if cae is None else dummy_function
+    debug_out = cae.dpo if cae else dummy_function
+    debug_out(f"    # executing at {os.getcwd()}: {args}")
+    pipe = None if lines_output is None else subprocess.PIPE
+    result: Union[subprocess.CompletedProcess, subprocess.CalledProcessError]   # having: stdout/stderr/returncode
+    try:
+        result = subprocess.run(args, stdout=pipe, stderr=pipe, input=console_input.encode(), check=True, shell=shell)
+    except subprocess.CalledProcessError as ex:                                             # pragma: no cover
+        debug_out(f"****  subprocess.run({args}) returned non-zero exit code {ex.returncode}; exception={ex}")
+        result = ex
+    except Exception as ex:
+        print_out(f"****  subprocess.run({args}) raised exception {ex}")
+        return 126
+
+    if lines_output is not None:
+        if result.stdout:
+            lines_output.extend([line for line in result.stdout.decode().split(os.linesep) if line])
+        if result.stderr:
+            lines_output.append("vvv   STDERR   vvv")
+            lines_output.extend([line for line in result.stderr.decode().split(os.linesep) if line])
+            lines_output.append("^^^   STDERR   ^^^")
+    return result.returncode
+
+
 class ConsoleApp(AppBase):
     """ provides command line arguments and options, config options, logging and debugging for your application.
 
@@ -344,6 +379,7 @@ class ConsoleApp(AppBase):
                                         supported kwargs are all the method kwargs of
                                         :meth:`~.core.AppBase.init_logging`.
         """
+        self._user_id = ''
         if not sys_env_id:
             sys_env_id = env_str(MAIN_SECTION_NAME + '_sys_env_id', convert_name=True) or ''
 
@@ -352,14 +388,14 @@ class ConsoleApp(AppBase):
 
         with config_lock:
             self._cfg_parser = instantiate_config_parser()                  #: ConfigParser instance
-            self.cfg_options: Dict[str, Literal] = dict()                   #: all config options
-            self.cfg_opt_choices: Dict[str, Iterable] = dict()              #: all valid config option choices
-            self.cfg_opt_eval_vars: dict = cfg_opt_eval_vars or dict()      #: app-specific vars for init of cfg options
+            self.cfg_options: Dict[str, Literal] = {}                       #: all config options
+            self.cfg_opt_choices: Dict[str, Iterable] = {}                  #: all valid config option choices
+            self.cfg_opt_eval_vars: dict = cfg_opt_eval_vars or {}          #: app-specific vars for init of cfg options
 
             # prepare config files, including default config file (last existing INI/CFG file) for
             # to write to. if there is no INI file at all then create on demand a <app_name>.INI file in the cwd.
             # note: the main INI file default file path will possibly be overwritten by :meth:`.load_cfg_files`.
-            self._cfg_files: list = list()                                  #: list of all found INI/CFG files
+            self._cfg_files: list = []                                      #: list of all found INI/CFG files
             self._main_cfg_fnam: str = os.path.join(os.getcwd(), self.app_name + INI_EXT)  #: def main config file name
             self._main_cfg_mod_time: float = 0.0                            #: main config file modification datetime
             warn_msg = self.add_cfg_files(*additional_cfg_files)
@@ -373,8 +409,7 @@ class ConsoleApp(AppBase):
             """
         self.load_cfg_files()
 
-        self.registered_users: Dict[str, Dict[str, Any]] = dict()
-        self._user_id = ''
+        self.registered_users: Dict[str, Dict[str, Any]] = {}
         self.user_specific_cfg_vars: Set[Tuple[str, str]] = set()
         self._init_default_user_cfg_vars()
         self.load_user_cfg()
@@ -416,7 +451,7 @@ class ConsoleApp(AppBase):
                                     or empty string if no logging got configured in cfg/args.
 
         the logging configuration can be specified in several alternative places. the precedence
-        on various existing configurations is (highest precedence first):
+        on various existing configurations is (the highest precedence first):
 
         * :ref:`log_file  <pre-defined-config-options>` :ref:`configuration option <config-options>` specifies
           the name of the used ae log file (will be read after initialisation of this app instance)
@@ -448,8 +483,8 @@ class ConsoleApp(AppBase):
                 log_file_name = self.get_var('log_file', default_value=logging_params.get('log_file_name'))
                 logging_params['log_file_name'] = log_file_name             # .. finally cfg log_file / log file arg
 
-        if logging_params.get('log_file_name'):                             # replace placeholders if has log file path
-            logging_params['log_file_name'] = norm_path(logging_params['log_file_name'])
+        if logging_params.get('log_file_name'):                             # if log file path: replace placeholders
+            logging_params['log_file_name'] = normalize(logging_params['log_file_name'])
 
         super().init_logging(**logging_params)
 
@@ -540,7 +575,7 @@ class ConsoleApp(AppBase):
             with config_lock:
                 return tuple((cfg_parser or self._cfg_parser).options(section))
         except NoSectionError:
-            self.dpo(f"ConsoleApp.cfg_section_variable_names: ignoring missing config file section {section}")
+            self.vpo(f"   ## ConsoleApp.cfg_section_variable_names: ignoring missing config file section {section}")
             return tuple()
 
     def _get_cfg_parser_val(self, name: str, section: str,
@@ -589,7 +624,7 @@ class ConsoleApp(AppBase):
                      cfg_parser: Optional[ConfigParser] = None, value_type: Optional[Type] = None) -> Any:
         """ determine value of a :ref:`config option <config-options>` or a :ref:`config variable <config-variables>`.
 
-        :param name:            id/name of a :ref:`config option <config-options>` or the name of a existing/declared
+        :param name:            id/name of a :ref:`config option <config-options>` or the name of an existing/declared
                                 :ref:`config variable <config-variables>`.
         :param section:         name of the :ref:`config section <config-sections>`. defaulting to the app options
                                 section (:data:`MAIN_SECTION_NAME`) if not specified or if None or empty string passed.
@@ -715,24 +750,21 @@ class ConsoleApp(AppBase):
     get_arg = get_argument      #: alias of method :meth:`.get_argument`
 
     def add_option(self, name: str, desc: str, value: Any,
-                   short_opt: str = None, choices: Optional[Iterable] = None, multiple: bool = False):
+                   short_opt: Union[str, UnsetType] = '', choices: Optional[Iterable] = None, multiple: bool = False):
         """ defining and adding a new config option for this app.
 
         :param name:        string specifying the option id and short description of this new option.
                             the name value will also be available as long command line argument option (case-sens.).
         :param desc:        description and command line help string of this new option.
-        :param value:       default value and the type of the option. the passed value will be used only if this option
+        :param value:       default value and type of the option. returned by :meth:`.get_option` if this option
                             is not specified as command line argument nor exists as config variable in any config file.
-                            the command line argument option value will always overwrite this value (and any value in
-                            any config file).
-
                             pass `UNSET` to define a boolean flag option, specified without a value on the command line.
                             the resulting value will be `True` if the option will be specified on the command line, else
                             `False`. specifying a value on the command line results in a `SystemExit` on parsing.
-
         :param short_opt:   short option character. if not passed or passed as '' then the first character of the name
-                            will be used. please note that the short options 'D' and 'L' are already used internally
-                            by :class:`ConsoleApp` (recommending using lower-case options for your application).
+                            will be used. passing `UNSET` or `None` prevents the declaration of a short option. please
+                            note that the short options 'h', 'D' and 'L' are already used internally by the classes
+                            :class:`~argparse.ArgumentParser` and :class:`ConsoleApp`.
         :param choices:     list of valid option values (optional, default=allow all values).
         :param multiple:    True if option can be added multiple times to command line (optional, default=False).
 
@@ -748,9 +780,12 @@ class ConsoleApp(AppBase):
         if short_opt == '':
             short_opt = name[0]
 
-        args = list()
+        args = []
         if short_opt and len(short_opt) == 1:
-            args.append('-' + short_opt)
+            short_opt = '-' + short_opt
+            # noinspection PyProtectedMember
+            assert short_opt not in self._arg_parser._option_string_actions, f"short_opt {short_opt} already exists"
+            args.append(short_opt)
         args.append('--' + name)
 
         # determine config value to use as default for command line arg
@@ -782,7 +817,7 @@ class ConsoleApp(AppBase):
             self.debug_level = value
 
     def get_option(self, name: str, default_value: Optional[Any] = None) -> Any:
-        """ determine the value of a config option specified by it's name (option id).
+        """ determine the value of a config option specified by its name (option id).
 
         :param name:            name/id of the config option.
         :param default_value:   default value of the option (if not defined with :class:`~ConsoleApp.add_option`).
@@ -862,12 +897,12 @@ class ConsoleApp(AppBase):
 
         if self.debug:
             debug_levels = ", ".join([str(k) + "=" + v for k, v in DEBUG_LEVELS.items()])
-            self.po(f"  ##  Debug Level({debug_levels}): {self.debug_level}", logger=_LOGGER)
+            self.po(f"  ##  debug level({debug_levels}): {self.debug_level}", logger=_LOGGER)
             if self._log_file_name:
-                self.po(f"   #  Log File: {self._log_file_name}", logger=_LOGGER)
+                self.po(f"   #  log file: {self._log_file_name}", logger=_LOGGER)
             if self.user_id:
-                self.po(f"   #  User Id: {self.user_id}", logger=_LOGGER)
-            self.po(f"  ##  {self.app_key} System Environment:", logger=_LOGGER)
+                self.po(f"   #  user id: {self.user_id}", logger=_LOGGER)
+            self.po(f"  ##  {self.app_key} system environment:", logger=_LOGGER)
             self.po(sys_env_text(extra_sys_env_dict=self.app_env_dict()), logger=_LOGGER)
 
     # app user related properties and methods
@@ -896,11 +931,11 @@ class ConsoleApp(AppBase):
                     usr_id = self._get_cfg_parser_val('user_id', MAIN_SECTION_NAME, default_value=os_user_name())
                 self.user_id = usr_id
 
-            reg_users = self.get_var('registered_users', default_value=dict())
+            reg_users = self.get_var('registered_users', default_value={})
             if reg_users:
                 self.registered_users = reg_users
 
-            usr_data = reg_users.get(self.user_id, dict())
+            usr_data = reg_users.get(self.user_id, {})
             self.user_specific_cfg_vars = usr_data.get('user_specific_cfg_vars',
                                                        self.get_var('user_specific_cfg_vars',
                                                                     default_value=self.user_specific_cfg_vars))
